import hashlib
import time

from pyramid.registry import global_registry


class TaskRegistry(object):
    _shared_state = {}
    _tasks = {}

    def __init__(self):
        self.__init__ = self._shared_state

    def register_task(self, func):
        self._tasks[func.__name__] = func


class Task(object):

    def __init__(self, func):
        self.func = func

    def get_task_id(self):
        timestamp = time.time()
        return hashlib.md5(str(timestamp).encode('utf-8')).hexdigest()

    def delay(self, *args, **kwargs):
        self.task_id = self.get_task_id()
        global_registry.producer.socket.send_pyobj([
            self.task_id,
            self.func.__name__,
            args,
            kwargs
        ])
        return self.task_id

    def __call__(self, settings, *args, **kwargs):
        self.func(settings, *args, **kwargs)

    @property
    def ready(self):
        cache = global_registry.result_cache
        return self.task_id in cache

    @property
    def result(self):
        cache = global_registry.result_cache
        return cache.get_value(self.task_id)

    def __repr__(self):
        return '<{}.{} for "{}" function>'.format(self.__module__, self.__class__.__name__, self.func.__name__)


def task(func):
    task = Task(func)
    registry = TaskRegistry()
    registry.register_task(func)
    return task



import unittest

from mock import patch
from pyramid import testing
from pyramid.response import Response

from papaye.tests.tools import (
    FakeGRequestResponse,
    FakeRoute,
    disable_cache,
    remove_blob_dir,
    set_database_connection,
)


class PackageTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.blob_dir = set_database_connection(self.request)
        settings = disable_cache()
        self.config = testing.setUp(request=self.request, settings=settings)

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    @patch('requests.get')
    def test_get_last_remote_version_without_proxy(self, mock):
        from papaye.models import Package

        fake_response = Response(status=200, body='{"info": {"version": "1.0"}}')
        mock.return_value = fake_response

        # Test data
        package = Package('test')

        result = Package.get_last_remote_version(False, package.name)

        self.assertEqual(mock.call_count, 0)
        self.assertIsNone(result)

    @patch('requests.get')
    def test_get_last_remote_version_with_proxy(self, mock):
        from papaye.models import Package

        fake_response = FakeGRequestResponse(status_code=200, content=b'{"info": {"version": "1.0"}}')
        mock.return_value = fake_response

        # Test data
        package = Package('test')

        result = Package.get_last_remote_version(True, package.name)

        self.assertEqual(mock.call_count, 1)
        self.assertIsNotNone(result)
        self.assertEqual(result, '1.0')

    @patch('requests.get')
    def test_get_last_remote_version_with_proxy_404(self, mock):
        from papaye.models import Package

        fake_response = FakeGRequestResponse(status_code=404, content=b'{"info": {"version": "1.0"}}')
        mock.return_value = fake_response

        # Test data
        package = Package('test')

        result = Package.get_last_remote_version(True, package.name)

        self.assertEqual(mock.call_count, 1)
        self.assertIsNone(result)

    def test_repository_is_up_to_date(self):
        from papaye.models import Package, Release

        # Test package
        package = Package(name='package1')
        release = Release(name='1.0', version='1.0')
        package['release'] = release

        self.assertTrue(package.repository_is_up_to_date('1.0'))
        self.assertTrue(package.repository_is_up_to_date('0.9'))
        self.assertTrue(package.repository_is_up_to_date('1.0a'))
        self.assertFalse(package.repository_is_up_to_date('1.1'))
        self.assertFalse(package.repository_is_up_to_date('1.1a'))
        self.assertTrue(package.repository_is_up_to_date(''))
        self.assertTrue(package.repository_is_up_to_date(None))

    def test_by_name(self):
        from papaye.models import Package
        from papaye.factories import repository_root_factory

        root = repository_root_factory(self.request)
        root['package1'] = Package(name='package1')

        result = Package.by_name('package1', self.request)
        self.assertEqual(result, root['package1'])

    def test_by_name_not_found(self):
        from papaye.models import Package

        result = Package.by_name('package1', self.request)
        self.assertEqual(result, None)

    def test_get_last_release(self):
        from papaye.models import Package, Release

        package = Package(name='package1')
        package.releases.update([('{}.0'.format(index), Release('', '{}.0'.format(index))) for index in range(1, 3)])
        result = package.get_last_release()
        self.assertEqual(result.version, '2.0')

    def test_get_last_release_with_minor(self):
        from papaye.models import Package, Release

        package = Package(name='package1')
        package.releases.update([('1.{}'.format(index), Release('', '1.{}'.format(index))) for index in range(1, 3)])
        result = package.get_last_release()
        self.assertEqual(result.version, '1.2')

    def test_get_last_release_with_alpha(self):
        from papaye.models import Package, Release

        package = Package(name='package1')
        package.releases.update([('1.0{}'.format(version), Release('', '1.0{}'.format(version)))
                                 for version in ['', 'a1', 'a2', 'b1', 'b2', 'rc1']])
        result = package.get_last_release()
        self.assertEqual(result.version, '1.0')


class ReleaseTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.config = testing.setUp(request=self.request)
        self.blob_dir = set_database_connection(self.request)

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    def test_by_packagename(self):
        from papaye.models import Release, Package
        from papaye.factories import repository_root_factory

        root = repository_root_factory(self.request)
        root['package1'] = Package(name='package1')
        root['package1']['1.0'] = Release('1.0', '1.0')

        result = Release.by_packagename('package1', self.request)
        self.assertEqual(result, [root['package1']['1.0'], ])

    def test_by_packagename_not_found(self):
        from papaye.models import Release

        result = Release.by_packagename('package1', self.request)
        self.assertEqual(result, None)


class UserTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.blob_dir = set_database_connection(self.request)
        self.config = testing.setUp(request=self.request)

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    def test_hash_password(self):
        from papaye.models import User
        expected = 'b109f3bbbc244eb82441917ed06d618b9008dd09b3befd1b5e07394c706a8bb980b1d7785e5976ec049b46df5f1326af5a'
        expected += '2ea6d103fd07c95385ffab0cacbc86'

        result = User('a_user', 'password')
        self.assertEqual(result.password, expected)

    def test_by_username(self):
        from papaye.models import User
        from papaye.factories import user_root_factory

        root = user_root_factory(self.request)
        root['a_user'] = User('a_user', 'password')

        result = User.by_username('a_user', self.request)

        self.assertIsInstance(result, User)
        self.assertEqual(result.username, 'a_user')

    def test_by_username_without_result(self):
        from papaye.models import User

        result = User.by_username('a_user', self.request)

        self.assertIsNone(result)



import shutil
import unittest

from mock import patch
from pyramid import testing
from repoze.evolution import ZODBEvolutionManager

from papaye.tests.tools import set_database_connection


class ConfigTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest()
        self.config = testing.setUp(request=self.request)
        self.blob_dir = set_database_connection(self.request, config=self.config)

    def tearDown(self):
        shutil.rmtree(self.blob_dir)

    def get_root(self):
        from papaye.models import Root
        root = Root()

        def root_func():
            return {'papaye_root': {'repository': {}, 'user': {}}}

        class P_jar:
            pass
        root._p_jar = P_jar()
        root._p_jar.root = root_func
        return root

    @patch('papaye.models.get_manager')
    def test_check_database_config(self, mock):
        from papaye import check_database_config
        root = self.get_root()
        mock.return_value = ZODBEvolutionManager(
            root,
            evolve_packagename='papaye.evolve',
            sw_version=0,
            initial_db_version=0
        )
        result = check_database_config(self.config)
        self.assertTrue(result)

    @patch('papaye.models.get_manager')
    def test_check_database_config_outdated(self, mock):
        from papaye import check_database_config
        root = self.get_root()
        manager = ZODBEvolutionManager(
            root,
            evolve_packagename='papaye.evolve',
            sw_version=1,
            initial_db_version=0
        )
        manager.sw_version = 1
        mock.return_value = manager

        from pyramid.config import ConfigurationError
        self.assertRaises(ConfigurationError, check_database_config, self.config)

    @patch('papaye.models.get_manager')
    def test_check_database_config_without_approot(self, mock):
        from papaye import check_database_config
        root = self.get_root()
        manager = ZODBEvolutionManager(
            root,
            evolve_packagename='papaye.evolve',
            sw_version=0,
            initial_db_version=0
        )
        mock.return_value = manager

        # Remove app_root
        del(self.config.registry._zodb_databases[''].open().root()['papaye_root'])
        import transaction
        transaction.commit()

        from pyramid.config import ConfigurationError
        self.assertRaises(ConfigurationError, check_database_config, self.config)



import sys

from pyramid.paster import bootstrap
from pyramid.config import Configurator
from repoze.evolution import evolve_to_latest

from papaye.models import get_manager


def main(*argv, **kwargs):
    config_file = sys.argv[-1]
    result = bootstrap(config_file)
    config = Configurator()
    config.registry = result['registry']
    manager = get_manager(config)
    if manager.get_db_version() >= manager.get_sw_version():
        sys.stdout.write('Your database is already up do date... ')
    else:
        sys.stdout.write('Your database need to be updated... ')
        evolve_to_latest(manager)
    sys.stdout.write('Done!\n')



import mock
import unittest

from requests import ConnectionError

from papaye.tests.tools import FakeGRequestResponse


class ExistsOnPyPIMixinTest(unittest.TestCase):

    from papaye.views.mixins import ExistsOnPyPIMixin

    class FakeView(ExistsOnPyPIMixin, object):
        pass

    @mock.patch('requests.get')
    def test_exists_on_pypi(self, mock):
        mock.return_value = FakeGRequestResponse(200, '')

        view = self.FakeView()

        result = view.exists_on_pypi('http://pypi.python.org/pypi/example')

        self.assertTrue(result)

    @mock.patch('requests.get')
    def test_exists_on_pypi_with_not_found(self, mock):
        mock.return_value = FakeGRequestResponse(404, '')

        view = self.FakeView()

        result = view.exists_on_pypi('http://pypi.python.org/pypi/example')

        self.assertFalse(result)

    @mock.patch('requests.get')
    def test_exists_on_pypi_with_other_status_code(self, mock):
        mock.return_value = FakeGRequestResponse(500, '')

        view = self.FakeView()

        result = view.exists_on_pypi('http://pypi.python.org/pypi/example')

        self.assertFalse(result)

    @mock.patch('requests.get')
    def test_exists_on_pypi_with_exceptions(self, mock):

        mock.side_effect = ConnectionError()

        view = self.FakeView()
        result = view.exists_on_pypi('http://pypi.python.org/pypi/example')
        self.assertFalse(result)



import hashlib
import logging
import requests
import transaction

from papaye.factories import repository_root_factory
from papaye.tasks import task
from papaye.proxy import PyPiProxy


logger = logging.getLogger(__name__)


def get_connection(config):
    conn = config.registry._zodb_databases[''].open()
    return conn


@task
def download_release_from_pypi(config, package_name, release_name):
    conn = get_connection(config)
    proxy = PyPiProxy(conn, package_name)
    package = proxy.build_repository(release_name=release_name)
    if not package:
        logger.error('Package {} not found on PYPI'.format(package_name))
    root = repository_root_factory(conn)
    for release_file in package[release_name].release_files.values():
        logger.info('Download file "{}"'.format(release_file.filename))
        release_file.set_content(requests.get(release_file.pypi_url).content)
        with release_file.content.open() as content:
            if hashlib.md5(content.read()).hexdigest() != release_file.md5_digest:
                raise IOError('md5 check error')
    root[package.name] = package
    try:
        transaction.commit()
    except:
        transaction.abort()



import io
import json
import types
import unittest

from cgi import FieldStorage
from mock import patch
from pyramid import testing
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response
from pyramid.threadlocal import get_current_registry

from papaye.tests.tools import (
    FakeGRequestResponse,
    FakeRoute,
    disable_cache,
    remove_blob_dir,
    set_database_connection,
)


class ListPackageViewTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.blob_dir = set_database_connection(self.request)
        self.config = testing.setUp(request=self.request)
        self.config.add_route('simple', '/simple/*traverse', factory='papaye.factories:repository_root_factory')
        registry = get_current_registry()
        registry.settings = disable_cache()

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    def test_list_packages(self):
        from papaye.views.simple import ListPackagesView
        from papaye.models import Package, Root

        # Test packages
        root = Root()
        root['package1'] = Package(name='package1')
        root['package2'] = Package(name='package2')

        view = ListPackagesView(root, self.request)
        response = view()

        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual([package.name for url, package in response['objects']], ['package1', 'package2'])

    def test_list_packages_without_package(self):
        from papaye.views.simple import ListPackagesView
        from papaye.models import Root

        # Test packages
        root = Root()

        view = ListPackagesView(root, self.request)
        response = view()
        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual([package.name for url, package in response['objects']], [])


class ListReleaseViewTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.blob_dir = set_database_connection(self.request)
        self.config = testing.setUp(request=self.request)
        self.config.add_route('simple', '/simple/*traverse', factory='papaye.factories:repository_root_factory')
        registry = get_current_registry()
        registry.settings = disable_cache()

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    def test_list_releases_files(self):
        from papaye.views.simple import ListReleaseFileView
        from papaye.models import Package, Release, Root, ReleaseFile

        # Test packages
        root = Root()
        root['package1'] = Package(name='package1')
        root['package1']['release1'] = Release(name='release1', version='1.0')
        root['package1']['release1'].__parent__ = root['package1']
        root['package1']['release1']['releasefile1.tar.gz'] = ReleaseFile(filename='releasefile1.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release1']['releasefile1.tar.gz'].__parent__ = root['package1']['release1']

        view = ListReleaseFileView(root['package1'], self.request)
        response = view()

        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual(
            [(url, release) for url, release in response['objects']],
            [('http://example.com/simple/package1/release1/releasefile1.tar.gz#md5=12345',
              root['package1']['release1']['releasefile1.tar.gz'])]
        )

    def test_list_releases_files_with_multiple_files(self):
        from papaye.views.simple import ListReleaseFileView
        from papaye.models import Package, Release, Root, ReleaseFile

        # Test packages
        root = Root()
        root['package1'] = Package(name='package1')
        root['package1']['release1'] = Release(name='release1', version='1.0')
        root['package1']['release1'].__parent__ = root['package1']
        root['package1']['release1']['releasefile1.tar.gz'] = ReleaseFile(filename='releasefile1.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release1']['releasefile1.tar.gz'].__parent__ = root['package1']['release1']
        root['package1']['release1']['releasefile2.tar.gz'] = ReleaseFile(filename='releasefile2.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release1']['releasefile2.tar.gz'].__parent__ = root['package1']['release1']

        view = ListReleaseFileView(root['package1'], self.request)
        response = view()

        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual(
            [(url, release) for url, release in response['objects']],
            [
                ('http://example.com/simple/package1/release1/releasefile1.tar.gz#md5=12345',
                 root['package1']['release1']['releasefile1.tar.gz']),
                ('http://example.com/simple/package1/release1/releasefile2.tar.gz#md5=12345',
                 root['package1']['release1']['releasefile2.tar.gz']),
            ],
        )

    def test_list_releases_files_with_multiple_release(self):
        from papaye.views.simple import ListReleaseFileView
        from papaye.models import Package, Release, Root, ReleaseFile

        # Test packages
        root = Root()
        root['package1'] = Package(name='package1')
        root['package1']['release1'] = Release(name='release1', version='1.0')
        root['package1']['release1'].__parent__ = root['package1']
        root['package1']['release2'] = Release(name='release2', version='2.0')
        root['package1']['release2'].__parent__ = root['package1']
        root['package1']['release1']['releasefile1.tar.gz'] = ReleaseFile(filename='releasefile1.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release1']['releasefile1.tar.gz'].__parent__ = root['package1']['release1']
        root['package1']['release2']['releasefile2.tar.gz'] = ReleaseFile(filename='releasefile2.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release2']['releasefile2.tar.gz'].__parent__ = root['package1']['release2']

        view = ListReleaseFileView(root['package1'], self.request)
        response = view()

        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual(
            [(url, release) for url, release in response['objects']],
            [
                ('http://example.com/simple/package1/release1/releasefile1.tar.gz#md5=12345',
                 root['package1']['release1']['releasefile1.tar.gz']),
                ('http://example.com/simple/package1/release2/releasefile2.tar.gz#md5=12345',
                 root['package1']['release2']['releasefile2.tar.gz']),
            ],
        )

    def test_list_releases_files_with_another_package(self):
        from papaye.views.simple import ListReleaseFileView
        from papaye.models import Package, Release, Root, ReleaseFile

        # Test packages
        root = Root()
        root['package1'] = Package(name='package1')
        root['package2'] = Package(name='package2')
        root['package1']['release1'] = Release(name='release1', version='1.0')
        root['package1']['release1'].__parent__ = root['package1']
        root['package2']['release2'] = Release(name='release2', version='2.0')
        root['package2']['release2'].__parent__ = root['package2']
        root['package1']['release1']['releasefile1.tar.gz'] = ReleaseFile(filename='releasefile1.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release1']['releasefile1.tar.gz'].__parent__ = root['package1']['release1']
        root['package2']['release2']['releasefile2.tar.gz'] = ReleaseFile(filename='releasefile2.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package2']['release2']['releasefile2.tar.gz'].__parent__ = root['package2']['release2']

        view = ListReleaseFileView(root['package1'], self.request)
        response = view()

        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual(
            [(url, release) for url, release in response['objects']],
            [
                ('http://example.com/simple/package1/release1/releasefile1.tar.gz#md5=12345',
                 root['package1']['release1']['releasefile1.tar.gz']),
            ],
        )

        view = ListReleaseFileView(root['package2'], self.request)
        response = view()

        self.assertIsInstance(response, dict)
        self.assertIn('objects', response)
        self.assertIsInstance(response['objects'], types.GeneratorType)
        self.assertEqual(
            [(url, release) for url, release in response['objects']],
            [
                ('http://example.com/simple/package2/release2/releasefile2.tar.gz#md5=12345',
                 root['package2']['release2']['releasefile2.tar.gz']),
            ],
        )

    @patch('requests.get')
    @patch('papaye.models.Package.get_last_remote_version')
    def test_list_releases_files_with_new_remotes_release(self, mock, requests_mock):
        from papaye.views.simple import ListReleaseFileView
        from papaye.models import Package, Release, Root, ReleaseFile

        mock.return_value = "3.0"

        # Test packages
        root = Root()
        root['package1'] = Package(name='package1')
        root['package1']['release1'] = Release(name='release1', version='1.0')
        root['package1']['release1'].__parent__ = root['package1']
        root['package1']['release1']['releasefile1.tar.gz'] = ReleaseFile(filename='releasefile1.tar.gz',
                                                                          content=b'',
                                                                          md5_digest='12345')
        root['package1']['release1']['releasefile1.tar.gz'].__parent__ = root['package1']['release1']

        self.request.matchdict['traverse'] = (root['package1'].__name__, root['package1']['release1'].__name__)
        self.request.registry.settings['papaye.proxy'] = 'true'
        view = ListReleaseFileView(root['package1'], self.request)
        pypi_result = {
            'info': {
                'name': 'package1',
            },
            'releases': {
                'release1': [{
                    'filename': 'releasefile1.tar.gz',
                    'url': 'http://example.com/',
                    'md5_digest': 'fake md5',
                }]
            }
        }
        requests_mock.return_value = FakeGRequestResponse(200, bytes(json.dumps(pypi_result), 'utf-8'))
        response = view()

        self.assertIsInstance(response, dict)


class DownloadReleaseViewTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.blob_dir = set_database_connection(self.request)
        settings = disable_cache()
        self.config = testing.setUp(request=self.request, settings=settings)
        self.config.add_route('simple', '/simple/*traverse', factory='papaye.factories:repository_root_factory')

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    @patch('requests.get')
    def test_download_release(self, mock):
        from papaye.views.simple import DownloadReleaseView
        from papaye.models import ReleaseFile, Release, Package

        package = Package(name='package')
        release = Release(name='1.0', version='1.0')
        release_file = ReleaseFile(filename='releasefile-1.0.tar.gz', content=b'Hello')
        release_file.content_type = 'text/plain'
        package['1.0'] = release
        package['1.0']['releasefile-1.0.tar.gz'] = release_file

        pypi_response = {
            'info': {
                'name': 'package',
                'version': '1.0'
            },
            'releases': {
                '1.0': [{
                    'filename': 'releasefile1.tar.gz',
                    'url': 'http://example.com/',
                    'md5_digest': 'fake md5',
                }]
            }
        }
        mock.return_value = FakeGRequestResponse(200, bytes(json.dumps(pypi_response), 'utf-8'))

        self.request.registry.settings['papaye.proxy'] = 'true'
        view = DownloadReleaseView(release_file, self.request)
        result = view()

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.body, b'Hello')
        self.assertEqual(result.content_type, 'text/plain')
        self.assertEqual(result.content_disposition, 'attachment; filename="releasefile-1.0.tar.gz"')

    @patch('requests.get')
    def test_download_release_with_old_release(self, mock_requests):
        from papaye.views.simple import DownloadReleaseView
        from papaye.models import ReleaseFile, Release, Package

        pypi_response = {
            'info': {
                'name': 'package',
                'version': '2.0',
            },
            'releases': {
                '2.0': [{
                    'filename': 'releasefile1.tar.gz',
                    'url': 'http://example.com/',
                    'md5_digest': 'fake md5',
                }]
            }
        }
        # mock_last_version.return_value = '2.0'
        mock_requests.return_value = FakeGRequestResponse(200, bytes(json.dumps(pypi_response), 'utf-8'))

        package = Package(name='package')
        release = Release(name='1.0', version='1.0')
        release_file = ReleaseFile(filename='releasefile-1.0.tar.gz', content=b'Hello')
        package['1.0'] = release
        package['1.0']['releasefile-1.0.tar.gz'] = release_file

        self.request.matchdict['traverse'] = (package.__name__, release.__name__, release_file.__name__)
        self.request.traversed = (package.__name__, release.__name__)

        view = DownloadReleaseView(release_file, self.request)
        view.proxy = True
        result = view()

        self.assertIsInstance(result, HTTPNotFound)


class UploadReleaseViewTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.config = testing.setUp(request=self.request)

    def test_upload_release(self):
        from papaye.models import Root, Package, Release, ReleaseFile
        from papaye.views.simple import UploadView

        # Create a fake test file
        uploaded_file = io.BytesIO()
        uploaded_file.write(b"content")
        storage = FieldStorage()
        storage.filename = 'foo.tar.gz'
        storage.file = uploaded_file

        self.request.POST = {
            "content": storage,
            "some_metadata": "Fake Metadata",
            "version": "1.0",
            "name": "my_package",
            ":action": "file_upload",
            "md5_digest": "Fake MD5"
        }
        root = Root()
        self.request.root = root
        view = UploadView(root, self.request)
        result = view()

        self.assertIsInstance(result, Response)
        self.assertEqual(result.status_int, 200)
        self.assertTrue('my_package' in root)
        self.assertIsInstance(root['my_package'], Package)
        self.assertTrue(root['my_package'].releases.get('1.0', False))
        self.assertIsInstance(root['my_package']['1.0'], Release)
        self.assertTrue(root['my_package']['1.0'].release_files.get('foo.tar.gz', b''))
        self.assertIsInstance(root['my_package']['1.0']['foo.tar.gz'], ReleaseFile)
        self.assertEqual(root['my_package']['1.0']['foo.tar.gz'].md5_digest, "Fake MD5")
        self.assertIsNotNone(root['my_package']['1.0'].metadata)
        self.assertIsInstance(root['my_package']['1.0'].metadata, dict)

    def test_upload_release_already_exists(self):
        from papaye.models import Root, Package, Release, ReleaseFile
        from papaye.views.simple import UploadView

        # Create a fake test file
        uploaded_file = io.BytesIO()
        uploaded_file.write(b"content")
        storage = FieldStorage()
        storage.filename = 'foo.tar.gz'
        storage.file = uploaded_file

        self.request.POST = {
            "content": storage,
            "some_metadata": "Fake Metadata",
            "version": "1.0",
            "name": "my_package",
            ":action": "file_upload",
        }
        root = Root()

        # Create initial release
        package = Package('my_package')
        package['1.0'] = Release('1.0', '1.0')
        package['1.0']['foo.tar.gz'] = ReleaseFile('foo.tar.gz', b'')
        root['my_package'] = package

        view = UploadView(root, self.request)
        result = view()

        self.assertIsInstance(result, Response)
        self.assertEqual(result.status_int, 409)

    def test_upload_release_multiple_releasefile(self):
        from papaye.models import Root, Package, Release, ReleaseFile
        from papaye.views.simple import UploadView

        # Create a fake test file
        uploaded_file = io.BytesIO()
        uploaded_file.write(b"content")
        storage = FieldStorage()
        storage.filename = 'foo.zip'
        storage.file = uploaded_file

        self.request.POST = {
            "content": storage,
            "some_metadata": "Fake Metadata",
            "version": "1.0",
            "name": "my_package",
            ":action": "file_upload",
        }
        root = Root()

        # Create initial release
        package = Package('my_package')
        package['1.0'] = Release('1.0', '1.0')
        package['1.0']['foo.tar.gz'] = ReleaseFile('foo.tar.gz', b'')
        root['my_package'] = package

        view = UploadView(root, self.request)
        result = view()

        self.assertIsInstance(result, Response)
        self.assertEqual(result.status_int, 200)
        self.assertTrue('my_package' in root)
        self.assertIsInstance(root['my_package'], Package)
        self.assertTrue(root['my_package'].releases.get('1.0', False))
        self.assertIsInstance(root['my_package']['1.0'], Release)
        self.assertTrue(root['my_package']['1.0'].release_files.get('foo.tar.gz', b''))
        self.assertIsInstance(root['my_package']['1.0']['foo.tar.gz'], ReleaseFile)
        self.assertTrue(root['my_package']['1.0'].release_files.get('foo.zip', b''))
        self.assertIsInstance(root['my_package']['1.0']['foo.zip'], ReleaseFile)


class ListReleaseFileByReleaseViewTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest(matched_route=FakeRoute('simple'))
        self.blob_dir = set_database_connection(self.request)
        self.config = testing.setUp(request=self.request)
        self.config.add_route('simple', '/simple/*traverse', factory='papaye.factories:repository_root_factory')
        registry = get_current_registry()
        registry.settings = {
            'cache.regions': 'pypi',
            'cache.enabled': 'false',
        }

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    def test_list_release(self):
        from papaye.views.simple import ListReleaseFileByReleaseView
        from papaye.models import Package, Release, ReleaseFile
        from papaye.factories import repository_root_factory

        # Initial data
        root = repository_root_factory(self.request)
        package = Package('my_package')
        package['1.0'] = Release('1.0', '1.0')
        package['1.0']['foo.tar.gz'] = ReleaseFile('foo.tar.gz', b'')
        root['my_package'] = package
        view = ListReleaseFileByReleaseView(package['1.0'], self.request)

        result = view()

        self.assertIsInstance(result, dict)
        self.assertIn('objects', result)
        self.assertEqual(list(result['objects']), [('http://example.com/simple/my_package/1.0/foo.tar.gz/',
                                                    package['1.0']['foo.tar.gz'])])

    def test_list_release_with_two_release(self):
        from papaye.views.simple import ListReleaseFileByReleaseView
        from papaye.models import Package, Release, ReleaseFile
        from papaye.factories import repository_root_factory

        # Initial data
        root = repository_root_factory(self.request)
        package = Package('my_package')
        package['1.0'] = Release('1.0', '1.0')
        package['1.0']['foo.tar.gz'] = ReleaseFile('foo.tar.gz', b'')
        package['2.0'] = Release('2.0', '2.0')
        package['2.0']['foo2.tar.gz'] = ReleaseFile('foo2.tar.gz', b'')
        root['my_package'] = package
        view1 = ListReleaseFileByReleaseView(package['1.0'], self.request)
        view2 = ListReleaseFileByReleaseView(package['2.0'], self.request)

        result1 = view1()
        result2 = view2()

        self.assertIsInstance(result1, dict)
        self.assertIn('objects', result1)
        self.assertEqual(list(result1['objects']), [('http://example.com/simple/my_package/1.0/foo.tar.gz/',
                                                     package['1.0']['foo.tar.gz'])])
        self.assertIsInstance(result2, dict)
        self.assertIn('objects', result2)
        self.assertEqual(list(result2['objects']), [('http://example.com/simple/my_package/2.0/foo2.tar.gz/',
                                                     package['2.0']['foo2.tar.gz'])])



import configparser
import os


def get_settings(inifile):
    parser = configparser.ConfigParser()
    current_dir = os.path.abspath(os.path.curdir)
    parser.read(inifile)
    return dict(parser.items('app:main', vars={"here": current_dir}))



import json
import unittest

from mock import patch
from pyramid import testing
from requests.exceptions import ConnectionError

from papaye.tests.tools import (
    FakeGRequestResponse,
    disable_cache,
    get_resource,
    remove_blob_dir,
    set_database_connection,
)


class ProxyTest(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest()
        self.blob_dir = set_database_connection(self.request)
        settings = disable_cache()
        self.config = testing.setUp(request=self.request, settings=settings)
        with open(get_resource('pyramid.json'), 'rb') as pyramid_json:
            self.pypi_response = FakeGRequestResponse(200, pyramid_json.read())

    def tearDown(self):
        remove_blob_dir(self.blob_dir)

    @patch('requests.get')
    def test_get_remote_informations(self, mock):
        from papaye.proxy import PyPiProxy
        mock.return_value = self.pypi_response
        url = "http://pypi.python.org/pypi/pyramid/json"

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.get_remote_informations(url)
        self.assertIsInstance(result, dict)
        self.assertEqual(result['info']['name'], 'pyramid')

    @patch('requests.get')
    def test_get_remote_informations_404(self, mock):
        from papaye.proxy import PyPiProxy
        mock.return_value = FakeGRequestResponse(404, b'')
        url = "http://pypi.python.org/pypi/pyramid/json"

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.get_remote_informations(url)
        self.assertIsNone(result)

    @patch('requests.get')
    def test_get_remote_informations_connection_error(self, mock):
        from papaye.proxy import PyPiProxy
        mock.side_effect = ConnectionError
        url = "http://pypi.python.org/pypi/pyramid/json"

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.get_remote_informations(url)
        self.assertIsNone(result)

    @patch('requests.get')
    def test_build_repository(self, mock):
        from papaye.proxy import PyPiProxy
        from papaye.models import Package, ReleaseFile
        mock.return_value = self.pypi_response
        info_dict = json.loads(self.pypi_response.content.decode('utf-8'))

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.build_repository()

        self.assertIsInstance(result, Package)
        for release in result.releases.values():
            self.assertIn(release.__name__, info_dict['releases'].keys())
        self.assertEqual(len(result.releases.keys()), len(info_dict['releases'].keys()))
        self.assertEqual(len(result['1.5'].release_files.keys()), 1)
        self.assertIsInstance(result['1.5'].release_files.values()[0], ReleaseFile)
        self.assertTrue(getattr(result['1.5'].release_files.values()[0], 'pypi_url', None))
        self.assertEqual(result['1.5'].release_files.values()[0].pypi_url,
                         "https://pypi.python.org/packages/source/p/pyramid/pyramid-1.5.tar.gz")
        self.assertIsNotNone(result['1.5'].metadata)
        self.assertIsInstance(result['1.5'].metadata, dict)

    @patch('requests.get')
    def test_build_repository_404_error(self, mock):
        from papaye.proxy import PyPiProxy
        mock.return_value = FakeGRequestResponse(404, b'')

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.build_repository()

        self.assertIsNone(result)

    @patch('requests.get')
    def test_smart_merge(self, mock):
        from papaye.proxy import PyPiProxy
        from papaye.factories import repository_root_factory
        from papaye.models import Package, Release, ReleaseFile
        mock.return_value = self.pypi_response
        root = repository_root_factory(self.request)

        # Existing releases
        root['pyramid'] = Package(name='pyramid')
        root['pyramid']['1.4'] = Release(name='1.4', version='1.4')
        root['pyramid']['1.4']['pyramid-1.4.tar.gz'] = ReleaseFile(
            filename='pyramid-1.4.tar.gz',
            content=b'',
            md5_digest='12345'
        )

        package = Package(name='pyramid')
        package['1.5'] = Release(name='1.5', version='1.5')
        package['1.5']['pyramid-1.5.tar.gz'] = ReleaseFile(
            filename='pyramid-1.5.tar.gz',
            content=b'',
            md5_digest='12345'
        )

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.smart_merge(root, package)

        self.assertEqual([key for key in result.releases.keys()], ['1.4', '1.5'])

    @patch('requests.get')
    def test_smart_merge_with_existing_release(self, mock):
        from papaye.proxy import PyPiProxy
        from papaye.factories import repository_root_factory
        from papaye.models import Package, Release, ReleaseFile
        mock.return_value = self.pypi_response
        root = repository_root_factory(self.request)

        # Existing releases
        package = Package(name='pyramid')
        package['1.5'] = Release(name='1.5', version='1.5')
        package['1.5']['pyramid-1.5.tar.gz'] = ReleaseFile(
            filename='pyramid-1.5.tar.gz',
            content=b'',
            md5_digest='12345'
        )

        package = Package(name='pyramid')
        package['1.5'] = Release(name='1.5', version='1.5')
        package['1.5']['pyramid-1.5.tar.gz'] = ReleaseFile(
            filename='pyramid-1.5.tar.gz',
            content=b'',
            md5_digest='12345'
        )

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.smart_merge(root, package)

        self.assertEqual([key for key in result.releases.keys()], ['1.5', ])
        self.assertEqual(result['1.5']['pyramid-1.5.tar.gz'].md5_digest, '12345')

    @patch('requests.get')
    def test_smart_merge_with_unknown_package(self, mock):
        from papaye.proxy import PyPiProxy
        from papaye.factories import repository_root_factory
        from papaye.models import Package, Release, ReleaseFile
        mock.return_value = self.pypi_response
        root = repository_root_factory(self.request)

        package = Package(name='pyramid')
        package['1.5'] = Release(name='1.5', version='1.5')
        package['1.5']['pyramid-1.5.tar.gz'] = ReleaseFile(
            filename='pyramid-1.5.tar.gz',
            content=b'',
            md5_digest='12345'
        )

        proxy = PyPiProxy(self.request, 'pyramid')
        result = proxy.smart_merge(root, package)

        self.assertEqual([key for key in result.releases.keys()], ['1.5', ])
        self.assertEqual(result, package)



import io
import shutil
import tempfile
import unittest

from mock import patch
from pyramid import testing
from pyramid.threadlocal import get_current_request

from papaye.tests.tools import (
    FakeGRequestResponse,
    disable_cache,
    get_db_connection,
    get_resource,
)


class TestDownloadTask(unittest.TestCase):

    def setUp(self):
        self.blob_dir = tempfile.mkdtemp()
        request = testing.DummyRequest()
        settings = disable_cache()
        self.config = testing.setUp(request=request, settings=settings)
        self.conn = get_db_connection(self.blob_dir)

    def tearDown(self):
        shutil.rmtree(self.blob_dir)

    @patch('requests.get')
    @patch('papaye.tasks.download.get_connection')
    def test_download_release_from_pypi(self, get_connection_mock, request_mock):
        from papaye.tasks.download import download_release_from_pypi
        from papaye.factories import repository_root_factory
        request = get_current_request()
        get_connection_mock.return_value = self.conn

        json_response = open(get_resource('pyramid.json'), 'rb')
        release_file_content = open(get_resource('pyramid-1.5.tar.gz'), 'rb')
        request_mock.side_effect = [
            FakeGRequestResponse(200, b'', 'http://pypi.python.org/simple/pyramid/'),
            FakeGRequestResponse(200, json_response.read()),
            FakeGRequestResponse(200, release_file_content.read()),
        ]

        download_release_from_pypi(request.registry.settings, 'pyramid', '1.5')
        self.assertEqual(request_mock.call_count, 3)
        root = repository_root_factory(self.conn)
        self.assertIn('pyramid', root)
        self.assertIn('1.5', root['pyramid'].releases)
        self.assertIn('pyramid-1.5.tar.gz', root['pyramid']['1.5'].release_files)
        release_file = root['pyramid']['1.5']['pyramid-1.5.tar.gz']
        self.assertEqual(release_file.md5_digest, "8747658dcbab709a9c491e43d3b0d58b")
        self.assertEqual(release_file.filename, "pyramid-1.5.tar.gz")
        self.assertEqual(list(root['pyramid'].releases.keys()), ['1.5', ])

    @patch('requests.get')
    @patch('papaye.tasks.download.get_connection')
    def test_download_release_from_pypi_with_bad_md5(self, get_connection_mock, request_mock):
        from papaye.tasks.download import download_release_from_pypi
        from papaye.factories import repository_root_factory
        request = get_current_request()
        get_connection_mock.return_value = self.conn

        json_response = open(get_resource('pyramid.json'), 'rb')
        release_file_content = io.BytesIO(b'corrupted_file')
        request_mock.side_effect = [
            FakeGRequestResponse(200, b'', 'http://pypi.python.org/simple/pyramid/'),
            FakeGRequestResponse(200, json_response.read()),
            FakeGRequestResponse(200, release_file_content.read()),
        ]

        self.assertRaises(IOError, download_release_from_pypi, request.registry.settings, 'pyramid', '1.5')
        self.assertEqual(request_mock.call_count, 3)
        root = repository_root_factory(self.conn)
        self.assertNotIn('pyramid', root)



import hashlib
import io
import json
import logging
import magic
import pkg_resources
import requests

from BTrees.OOBTree import OOBTree
from ZODB.blob import Blob
from beaker.cache import cache_region
from persistent import Persistent
from pkg_resources import parse_version
from pyramid.security import Allow, ALL_PERMISSIONS, Everyone
from pyramid.threadlocal import get_current_registry
from pyramid_zodbconn import db_from_uri
from repoze.evolution import ZODBEvolutionManager
from requests.exceptions import ConnectionError

from papaye.factories import user_root_factory, repository_root_factory, APP_ROOT_NAME


logger = logging.getLogger(__name__)
SW_VERSION = 0


def get_manager(config):
    conn = config.registry._zodb_databases[''].open()
    root = conn.root()[APP_ROOT_NAME]
    manager = ZODBEvolutionManager(
        root,
        evolve_packagename='papaye.evolve',
        sw_version=SW_VERSION,
        initial_db_version=0
    )
    return manager


def format_key(key):
    return pkg_resources.safe_name(key.lower())


def get_connection(settings):
    uri = settings.get('zodbconn.uri', None)
    db = db_from_uri(uri, 'unamed', None)
    return db.open()


class Root(OOBTree):
    __name__ = __parent__ = None

    def __acl__(self):
        acl = [
            (Allow, 'group:installer', 'install'),
            (Allow, 'group:admin', ALL_PERMISSIONS)
        ]
        registry = get_current_registry()
        anonymous_install = registry.settings.get('papaye.anonymous_install')
        anonymous_install = True if anonymous_install and anonymous_install == 'true' else False
        if anonymous_install:
            acl.append((Allow, Everyone, 'install'))
        return acl

    def __getitem__(self, package_name):
        keys = [key for key in self.keys()
                if format_key(key) == format_key(package_name)]
        if len(keys) == 1:
            return self.get(keys[0])


class Package(Persistent):
    pypi_url = 'http://pypi.python.org/pypi/{}/json'

    def __init__(self, name):
        self.__name__ = name
        self.name = name
        self.releases = OOBTree()

    def __getitem__(self, release_name):
        return self.releases[release_name]

    def __setitem__(self, key, value):
        key = format_key(key)
        self.releases[key] = value
        self.releases[key].__parent__ = self

    @classmethod
    @cache_region('pypi', 'get_last_remote_filename')
    def get_last_remote_version(cls, proxy, package_name):
        logger.debug('Not in cache')
        if not proxy:
            return None
        try:
            result = requests.get('http://pypi.python.org/pypi/{}/json'.format(package_name))
            if not result.status_code == 200:
                return None
            result = json.loads(result.content.decode('utf-8'))
            return result['info']['version']
        except ConnectionError:
            pass
        return None

    def repository_is_up_to_date(self, last_remote_release):
        if not last_remote_release:
            return True
        remote_version = parse_version(last_remote_release)

        local_versions = [release.version for release in self.releases.values()]
        for version in local_versions:
            if parse_version(version) >= remote_version:
                return True
        return False

    @classmethod
    def by_name(cls, name, request):
        root = repository_root_factory(request)
        return root[name] if name in root else None

    def get_last_release(self):
        max_version = max([parse_version(version) for version in self.releases.keys()])
        for version, release in self.releases.items():
            if parse_version(version) == max_version:
                return release


class Release(Persistent):

    def __init__(self, name, version, metadata=None):
        self.__name__ = name
        self.release_files = OOBTree()
        self.version = version
        self.metadata = metadata

    def __getitem__(self, release_file_name):
        return self.release_files[format_key(release_file_name)]

    def __setitem__(self, key, value):
        key = format_key(key)
        self.release_files[key] = value
        self.release_files[key].__parent__ = self

    @classmethod
    def by_packagename(cls, package, request):
        root = repository_root_factory(request)
        if package not in root:
            return None
        return list(root[package].releases.values())


class ReleaseFile(Persistent):

    def __init__(self, filename, content, md5_digest=None):
        self.filename = self.__name__ = filename
        self.md5_digest = md5_digest
        self.set_content(content)

    def get_content_type(self, content):
        buf = io.BytesIO(content)
        with magic.Magic(flags=magic.MAGIC_MIME_TYPE) as m:
            self.content_type = m.id_buffer(buf.read())

    def set_content(self, content):
        self.content = Blob(content)
        self.content_type = self.get_content_type(content)


class User(Persistent):

    def __init__(self, username, password, **kwargs):
        self.username = username
        self.password = self.hash_password(password)
        self.groups = kwargs.get('groups', [])

    def hash_password(self, password):
        return hashlib.sha512(password.encode('utf-8')).hexdigest()

    def password_verify(self, clear_password):
        return self.hash_password(clear_password) == self.password

    @classmethod
    def by_username(cls, username, request):
        """Return user instance by username"""
        root = user_root_factory(request)
        if username not in root:
            return None
        else:
            return root[username]



import logging
import requests

from beaker.cache import cache_region


logger = logging.getLogger(__name__)


class ExistsOnPyPIMixin(object):

    @cache_region('pypi', 'get_last_remote_filename')
    def exists_on_pypi(self, pypi_url):
        try:
            result = requests.get(pypi_url)
            if result.status_code == 200:
                return result
            else:
                return None
        except:
            logger.info('Bad url "{}"'.format(pypi_url))
            return None



import os
import shutil
import tempfile

from pyramid import testing
from pyramid_beaker import set_cache_regions_from_settings
from ZODB.blob import BlobStorage
from ZODB.DB import DB
from ZODB.MappingStorage import MappingStorage

from papaye.scripts.initialize import create_app_root


HERE = os.path.abspath(os.path.dirname(__name__))
TEST_RESOURCES = os.path.join('..', '..', HERE, 'test-resources')


def get_resource(*args):
    return os.path.join(TEST_RESOURCES, *args)


class FakeGRequestResponse():

    def __init__(self, status_code, content, url="https://fake/url/"):
        self.status_code = status_code
        self.content = content
        self.url = url


class FakeMatchedDict(object):

    def __init__(self, name):
        self.name = name


class TestRequest(testing.DummyRequest):

    def __init__(self, *args, **kwargs):
        super(TestRequest, self).__init__(*args, **kwargs)
        self.matched_route = FakeMatchedDict('')


class FakeRoute():

    def __init__(self, name):
        self.name = name


class FakeProducer(object):

    def __init__(self, *args, **kwargs):
        pass

    def send_pyobj(self, val):
        pass


def create_test_app(config):
    """ This function returns a Pyramid WSGI application for functional tests.
    """
    config.include('pyramid_jinja2')
    set_cache_regions_from_settings(config.registry.settings)
    config.add_jinja2_search_path("papaye:templates")
    config.add_static_view('static', 'static', cache_max_age=3600)
    config.add_route('simple', '/simple/*traverse', factory='papaye:root_factory')
    config.scan()
    return config.make_wsgi_app()


def get_db_connection(blob_dir):
    storage = MappingStorage('test')
    blob_storage = BlobStorage(blob_dir, storage)
    db = DB(blob_storage)
    conn = db.open()
    create_app_root(conn)
    return conn


def set_database_connection(request, blob_dir=None, config=None):
    if not blob_dir:
        blob_dir = tempfile.mkdtemp('blobs')
    conn = get_db_connection(blob_dir)
    request._primary_zodb_conn = conn
    if config:
        import transaction
        conn.root()['repository'] = {}
        conn.root()['user'] = {}
        transaction.commit()
        config.registry._zodb_databases = {'': conn._db}
    return blob_dir


def remove_blob_dir(blob_dir):
    shutil.rmtree(blob_dir)


def disable_cache(settings=None):
    cache_settings = {
        'cache.regions': 'pypi',
        'cache.enabled': 'false',
    }
    if settings:
        settings.update(cache_settings)
    else:
        settings = cache_settings
    set_cache_regions_from_settings(settings)
    return settings



import logging
import multiprocessing
import pickle
import signal
import traceback
import zmq

# from beaker.cache import CacheManager
# from beaker.util import parse_cache_config_options
# from pyramid.registry import global_registry

from papaye.tasks import TaskRegistry


logger = logging.getLogger(__name__)


class Device(multiprocessing.Process):

    def __init__(self, config):
        super().__init__()
        self.config = config
        self.settings = self.config.registry.settings
        self.go = True

    def get_socket(self):
        raise NotImplemented()

    def run(self):
        raise NotImplemented()


class Scheduler(object):

    def __init__(self, settings, config):
        self.config = config
        self.settings = settings
        self.concurency = int(self.settings.get('papaye.worker.concurency', 1))
        signal.signal(signal.SIGTERM, self.sigterm_handler)
        self.devices = []

    def sigterm_handler(self, _signo, _stack_frame):
        for device in self.devices:
            if device._popen:
                device.terminate()

    def run(self):
        self.devices.append(QueueDevice(self.config))
        # self.devices.append(CollectorDevice(self.config))
        for index in range(1, self.concurency + 1):
            self.devices.append(ConsumerDevice(self.config, index))
        for device in self.devices:
            device.daemon = True
            device.start()


class Producer(object):

    def __init__(self, config):
        self.config = config
        self.settings = self.config.registry.settings
        self.socket = self.get_socket()

    def get_socket(self):
        context = zmq.Context()
        socket = context.socket(zmq.XREQ)
        socket.connect(self.settings.get('papaye.proxy.broker'))
        return socket


class ConsumerDevice(Device):

    def __init__(self, config, worker_number):
        super().__init__(config)
        self.worker_number = worker_number
        # Load tasks
        __import__("papaye.tasks.download")

    def get_sockets(self):
        context = zmq.Context()
        socket = context.socket(zmq.XREP)
        socket.connect(self.settings.get('papaye.proxy.worker_socket'))
        socket2 = context.socket(zmq.PUSH)
        socket2.connect(self.settings.get('papaye.proxy.collector_socket'))
        return socket, socket2

    def run(self):
        logger.info('Starting worker {}'.format(self._name))
        self.worker_socket, self.collector_socket = self.get_sockets()
        worker_number = self.worker_number
        while self.go:
            try:
                data = self.worker_socket.recv_multipart()[2]
                task_id, func_name, args, kwargs = pickle.loads(data)
                func = TaskRegistry()._tasks[func_name]
                func.task_id = task_id
                logger.info('Worker {}: Starting task id: {}'.format(worker_number, func.task_id))
                try:
                    # result = func(self.settings, *args, **kwargs)
                    # result = func(self.config, *args, **kwargs)
                    func(self.config, *args, **kwargs)
                    # self.collector_socket.send(pickle.dumps((task_id, result)))
                    logger.info('Worker {}: Task #{} finished'.format(worker_number, func.task_id))
                except Exception as exc:
                    formated_tb = traceback.format_tb(exc.__traceback__)
                    logger.error('Worker {}: Task #{} Error\n{}'.format(worker_number, func.task_id, formated_tb))
            except KeyboardInterrupt:
                self.go = False


class QueueDevice(Device):

    def __init__(self, config):
        super().__init__(config)
        # self.frontend, self.backend = self.get_sockets()

    def get_sockets(self):
        context = zmq.Context(1)
        # Socket facing clients
        frontend = context.socket(zmq.XREP)
        frontend.bind(self.settings.get('papaye.proxy.broker'))
        # Socket facing services
        backend = context.socket(zmq.XREQ)
        backend.bind(self.settings.get('papaye.proxy.worker_socket'))
        return frontend, backend

    def run(self):
        logger.debug('Starting {} device'.format(self._name))
        try:
            self.frontend, self.backend = self.get_sockets()
            zmq.device(zmq.QUEUE, self.frontend, self.backend)
        except KeyboardInterrupt:
            pass


# class CollectorDevice(Device):

#     def __init__(self, config):
#         super().__init__(config)
#         cache_manager = CacheManager(**parse_cache_config_options(self.settings))
#         self.cache = cache_manager.get_cache_region('result_cache', 'result')
#         global_registry.result_cache = self.cache

#     def get_socket(self):
#         context = zmq.Context()
#         socket = context.socket(zmq.PULL)
#         socket.bind(self.settings.get('proxy.collector_socket'))
#         return socket

#     def run(self):
#         socket = self.get_socket()
#         while self.go:
#             try:
#                 data = socket.recv()
#                 task_id, value = pickle.loads(data)
#                 if value:
#                     self.cache.set_value(task_id, value)
#                 else:
#                     self.cache.set_value(task_id, 2)
#             except KeyboardInterrupt:
#                 self.go = False



import getpass
import sys
import transaction

from papaye.scripts.common import get_settings
from papaye.factories import user_root_factory, APP_NAME
from papaye.models import User, Root, get_connection


def get_database_root(dbconn):
    zodb_root = dbconn.root()
    return zodb_root


def app_root_exists(dbconn):
    zodb_root = get_database_root(dbconn)
    if not '{}_root'.format(APP_NAME) in zodb_root:
        return False
    return True


def create_app_root(dbconn):
    zodb_root = get_database_root(dbconn)
    if not '{}_root'.format(APP_NAME) in zodb_root:
        zodb_root['{}_root'.format(APP_NAME)] = Root()
    if not 'user' in zodb_root['{}_root'.format(APP_NAME)]:
        zodb_root['{}_root'.format(APP_NAME)]['user'] = Root()
    if not 'repository' in zodb_root['{}_root'.format(APP_NAME)]:
        zodb_root['{}_root'.format(APP_NAME)]['repository'] = Root()


def create_admin_user(dbconn):
    username = input('username for administrator (default="admin"): ')
    username = 'admin' if not username or username == '' else username
    password = getpass.getpass()
    admin = User(username, password, groups=['group:admin'])
    root = user_root_factory(dbconn)
    root[username] = admin
    transaction.commit()


def main(*argv, **kwargs):
    settings = get_settings(sys.argv[1])
    conn = get_connection(settings)
    if not app_root_exists(conn):
        create_app_root(conn)
    create_admin_user(conn)
    print("Initialization complete!")



from pyramid_zodbconn import get_connection
from pyramid.request import Request
from pyramid.testing import DummyRequest


APP_NAME = __import__(__name__).__name__
APP_ROOT_NAME = '{}_root'.format(APP_NAME)


def user_root_factory(request_or_connection):
    if isinstance(request_or_connection, (Request, DummyRequest)):
        conn = get_connection(request_or_connection)
    else:
        conn = request_or_connection
    zodb_root = conn.root()
    if not '{}_root'.format(APP_NAME) in zodb_root:
        return None
    return zodb_root[APP_ROOT_NAME]['user']


def repository_root_factory(request_or_connection):
    if isinstance(request_or_connection, (Request, DummyRequest)):
        conn = get_connection(request_or_connection)
    else:
        conn = request_or_connection
    zodb_root = conn.root()
    if not '{}_root'.format(APP_NAME) in zodb_root:
        return None
    return zodb_root[APP_ROOT_NAME]['repository']



import copy
import json
import requests

from beaker.cache import cache_region
from requests.exceptions import ConnectionError

from papaye.factories import repository_root_factory
from papaye.models import Package, Release, ReleaseFile, Root


class PyPiProxy:
    pypi_simple_url = 'https://pypi.python.org/simple/{}/'
    pypi_url = 'https://pypi.python.org/pypi/{}/json'

    def __init__(self, request_or_dbconn,  package_name):
        self.request_or_dbconn = request_or_dbconn
        self.package_name = package_name

    @cache_region('pypi', 'get_remote_package_name')
    def get_remote_package_name(self, package_name):
        response = requests.get(self.pypi_simple_url.format(package_name))
        if response.status_code == 200:
            if response.url.endswith('/'):
                return response.url.split('/')[-2]
            else:
                return response.url.split('/')[-1]
        else:
            None

    @cache_region('pypi', 'get_remote_informations')
    def get_remote_informations(self, url):
        try:
            response = requests.get(url)
            if response.status_code == 200:
                return json.loads(response.content.decode('utf-8'))
            else:
                None
        except ConnectionError:
            return None

    def build_repository(self, release_name=None):
        root = repository_root_factory(self.request_or_dbconn)
        package_name = self.get_remote_package_name(self.package_name)
        if not package_name:
            return None
        info = self.get_remote_informations(self.pypi_url.format(package_name))
        if info:
            package_root = Root()
            package = Package(info['info']['name'])
            package.__parent__ = package_root
            package_root[package.name] = package

            remote_releases = [release_name, ] if release_name else info['releases'].keys()

            for remote_release in remote_releases:
                release = Release(remote_release, remote_release, metadata=info['info'])
                package[remote_release] = release

                for remote_release_file in info['releases'][remote_release]:
                    filename = remote_release_file['filename']
                    md5_digest = remote_release_file['md5_digest']
                    release_file = ReleaseFile(filename, b'', md5_digest)
                    setattr(release_file, 'pypi_url', remote_release_file['url'])
                    release[filename] = release_file
            return self.smart_merge(root, package)
        return None

    def smart_merge(self, root, package):
        if package.name not in root:
            return package
        else:
            merged_package = copy.deepcopy(root[package.name])
            to_delete_releases_name = [release for release in package.releases.keys()
                                       if release in merged_package.releases.keys()]
            to_update_releases = [(key, value) for key, value in package.releases.items()
                                  if key not in to_delete_releases_name]
            merged_package.releases.update(to_update_releases)
            return merged_package



# -*- coding:utf-8 -*-
import os

from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.rst')).read()
CHANGES = open(os.path.join(HERE, 'CHANGES.txt')).read()
VERSION = '0.1'

requires = [
    'filemagic',
    'pyramid',
    'pyramid_beaker',
    'pyramid_debugtoolbar',
    'pyramid_jinja2',
    'pyramid_tm',
    'pyramid_zodbconn',
    'pyzmq',
    'repoze.evolution',
    'requests',
    'transaction',
    'waitress',
]

setup(name='papaye',
      version=VERSION,
      description='papaye',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
      ],
      author='Romain Commandé',
      author_email='commande.romain@gmail.com',
      url='http://www.rcommande.org',
      keywords='web wsgi bfg pylons pyramid',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='papaye',
      install_requires=requires,
      entry_points="""\
      [paste.app_factory]
      main = papaye:main
      [console_scripts]
      papaye_init = papaye.scripts.initialize:main
      papaye_evolve = papaye.scripts.evolve:main
      # papaye_worker = papaye.scripts.worker:main
      papaye = papaye.scripts.papaye:main
      """,
      )



import logging


logger = logging.getLogger(__name__)


class BaseView(object):

    def __init__(self, context, request):
        logger.info('Dispatch "{}" route as {} with {} method'.format(
            request.matched_route.name,
            request.path,
            request.method
        ))
        self.request = request
        self.context = context
        self.settings = request.registry.settings
        proxy = self.settings.get('papaye.proxy', False)
        self.proxy = True if proxy and proxy == 'true' else False



import sys

from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator, ConfigurationError
from pyramid.httpexceptions import HTTPNotFound
from pyramid_beaker import set_cache_regions_from_settings

from papaye.factories import repository_root_factory, user_root_factory
from papaye.models import User
from papaye.tasks.devices import Scheduler, Producer
from pyramid.threadlocal import get_current_registry


def auth_check_func(username, password, request):
    user = User.by_username(username, request)
    if user and user.password_verify(password):
        return user.groups
    return None


def check_database_config(config):
    from papaye.models import get_manager
    manager = get_manager(config)
    if manager.get_db_version() < manager.get_sw_version():
        raise ConfigurationError('Your database need to be updated! Run '
                                 '"papaye_evolve path_to_your_config_file.ini" command first')
    conn = config.registry._zodb_databases[''].open()
    if user_root_factory(conn) is None or repository_root_factory(conn) is None:
        raise ConfigurationError('Database does not exist! Run "papaye_init '
                                 'path_to_your_config_file.ini command first')
    return True


def notfound(request):
    return HTTPNotFound()


def add_directives(config):
    config.add_directive('check_database_config', check_database_config)
    config.add_directive('start_scheduler', start_scheduler)


def start_scheduler(config):
    registry = get_current_registry()
    registry.producer = Producer(config)
    if sys.argv[0].endswith('pserve'):
        scheduler = Scheduler(config.registry.settings, config)
        scheduler.run()


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    set_cache_regions_from_settings(settings)
    config = Configurator(settings=settings)
    add_directives(config)
    authn_policy = BasicAuthAuthenticationPolicy(check=auth_check_func)
    authz_policy = ACLAuthorizationPolicy()
    config.set_authentication_policy(authn_policy)
    config.set_authorization_policy(authz_policy)
    config.add_jinja2_search_path("papaye:templates")
    config.add_static_view('static', 'static', cache_max_age=3600)
    config.add_route('home', '/')
    config.add_route('simple', '/simple*traverse', factory=repository_root_factory)
    config.add_notfound_view(notfound, append_slash=True)
    # check_database_config(config)
    config.check_database_config()
    config.scan()
    config.start_scheduler()
    return config.make_wsgi_app()



import logging

from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPConflict,
    HTTPNotFound,
    HTTPUnauthorized,
    HTTPTemporaryRedirect
)
from pyramid.response import Response
from pyramid.security import forget
from pyramid.view import view_config, notfound_view_config, forbidden_view_config

from papaye.models import ReleaseFile, Package, Release, Root
from papaye.proxy import PyPiProxy
from papaye.views.commons import BaseView
from papaye.tasks.download import download_release_from_pypi


logger = logging.getLogger(__name__)


@forbidden_view_config()
def basic_challenge(request):
    response = HTTPUnauthorized()
    response.headers.update(forget(request))
    return response


@notfound_view_config(route_name="simple", renderer='simple.jinja2')
def not_found(request):
    settings = request.registry.settings
    proxy = settings.get('papaye.proxy', False)
    proxy = True if proxy and proxy == 'true' else False
    if not proxy:
        return HTTPNotFound()
    try:
        proxy = PyPiProxy(request, request.matchdict['traverse'][0])
        package = proxy.build_repository()
        if not package:
            return HTTPNotFound()
        if len(request.matchdict['traverse']) == 1:
            view = ListReleaseFileView(package, request)
        elif len(request.matchdict['traverse']) == 2:
            context = package[request.matchdict['traverse'][1]]
            view = ListReleaseFileByReleaseView(context, request)
        elif len(request.matchdict['traverse']) == 3:
            release_file = package[request.matchdict['traverse'][1]][request.matchdict['traverse'][2]]
            package_name, release_name, _ = request.matchdict['traverse']
            download_release_from_pypi.delay(package_name, release_name)
            return HTTPTemporaryRedirect(location=release_file.pypi_url)
    except KeyError:
        return HTTPNotFound()
    return view()


@view_config(
    context="papaye.models.Root",
    route_name="simple",
    renderer="simple.jinja2",
    request_method="GET",
    permission="install"
)
class ListPackagesView(BaseView):

    def __call__(self):
        return {
            'objects': ((self.request.resource_url(package, route_name=self.request.matched_route.name), package)
                        for package in self.context.values()),
        }


@view_config(
    context=Package,
    route_name="simple",
    renderer="simple.jinja2",
    request_method="GET",
    permission='install'
)
class ListReleaseFileView(BaseView):

    def __call__(self):
        package = self.context
        release_files = []
        for release in package.releases.values():
            for release_file in release.release_files.values():
                release_files.append(release_file)
        context = {
            'objects': ((self.request.resource_url(
                release_file,
                route_name='simple'
            )[:-1] + "#md5={}".format(release_file.md5_digest), release_file) for release_file in release_files),
        }
        if package.repository_is_up_to_date(Package.get_last_remote_version(self.proxy, package.name)):
            return context
        else:
            return not_found(self.request)


@view_config(
    context=Release,
    route_name="simple",
    renderer="simple.jinja2",
    request_method="GET",
    permission='install'
)
class ListReleaseFileByReleaseView(BaseView):

    def __call__(self):
        return {
            'objects': (
                (self.request.resource_url(release_file, route_name=self.request.matched_route.name), release_file)
                for release_file in self.context.release_files.values()),
        }


@view_config(context=ReleaseFile, route_name="simple", permission='install', request_method="GET")
class DownloadReleaseView(BaseView):

    def __call__(self):
        check_update = True if self.request.GET.get('check_update', 'true') == 'true' else False
        package = self.context.__parent__.__parent__
        last_remote_version = Package.get_last_remote_version(self.proxy, package.name)
        if check_update:
            if not package.repository_is_up_to_date(last_remote_version):
                return not_found(self.request)
        response = Response()
        response.content_disposition = 'attachment; filename="{}"'.format(self.context.filename)
        response.charset = 'utf-8'
        response.content_type = self.context.content_type
        response.body_file = self.context.content.open()
        return response


@view_config(context=Root, route_name="simple", request_method="POST", permission="add")
class UploadView():

    def __init__(self, context, request):
        self.request = request
        self.context = context

    def __call__(self):
        post = dict(self.request.POST)
        metadata = dict([(key, value) for key, value in post.items() if key != 'content'])
        if post.get(':action') == "file_upload":
            name = post.get('name')
            version = post.get('version')
            content = post.get('content')
            md5_digest = post.get('md5_digest')

            package = self.context[name] if self.context.get(name) else Package(name)
            package.__parent__ = self.context
            self.context[name] = package

            release = package[version] if package.releases.get(version) else Release(name=version,
                                                                                     version=version,
                                                                                     metadata=metadata)
            release.__parent__ = package
            self.context[name][version] = release

            if release.release_files.get(content.filename):
                return HTTPConflict()

            release_file = ReleaseFile(filename=content.filename, content=content.file.read(), md5_digest=md5_digest)
            release = self.context[name][version]
            self.context[name][version][content.filename] = release_file
            return Response()
        else:
            return HTTPBadRequest()
